-- SAMPLE CODE for the Advanced CTE Presentation
-- Created by Steve Stedman
--    http://SteveStedman.com
--    twitter:   @SqlEmt

RAISERROR ('Dont run the whole script, just run sections at a time', 20, 1)  WITH LOG;

use [cte_demo];  -- download the cte_demo database from http://stevestedman.com/?p=2839
-- For more info on recursive CTE formatting see http://stevestedman.com/?p=1130
GO
-- hierarchy


-- To implement this lets start with the original departmentsCTE query from the last demo:
;WITH DepartmentCTE AS
(
	SELECT id, department, parent, 1 AS lvl 
  	  FROM Departments
	 WHERE parent IS NULL

	 UNION ALL 

	SELECT d.id, d.department, d.parent, dCTE.lvl + 1 AS lvl 
	  FROM Departments AS d 
	 INNER JOIN DepartmentCTE AS dCTE ON d.parent = dCTE.id
	 WHERE d.parent IS NOT NULL
) 
SELECT * 
  FROM DepartmentCTE;


EXEC dbo.FillMoreDepartments;

-- For the anchor query the full path (TreePath) is just the current department so 
--   we add a new column into the anchor query:

cast(department as varchar(8000)) as TreePath

-- For the recursive query we add a reference to the existing TreePath, followed by 
--   a separator and then the current department.

cast(dCTE.TreePath 
   + ' -> ' 
   + d.Department as varchar(8000)) as TreePath





;WITH DepartmentCTE AS
(
	SELECT id, Department, parent, 1 AS lvl,
           cast(Department as varchar(8000)) as TreePath 
  	  FROM Departments
	 WHERE parent IS NULL

	 UNION ALL 

	SELECT d.id, d.Department, d.parent, dCTE.lvl + 1 AS lvl,
		   cast(dCTE.TreePath 
                + ' -> ' 
                + d.Department AS varchar(8000)) as TreePath
	  FROM Departments AS d 
	 INNER JOIN DepartmentCTE AS dCTE ON d.parent = dCTE.id
	 WHERE d.parent IS NOT NULL
) 
SELECT * 
  FROM DepartmentCTE
 ORDER BY parent;








-- To improve even further on the query we can just show the TreePath column 
--   and order it by TreePath for a better visual representation.
;WITH DepartmentCTE AS
(
	SELECT id, Department, parent, 1 AS lvl,
           cast(Department as varchar(8000)) as TreePath 
  	  FROM Departments
	 WHERE parent IS NULL

	 UNION ALL 

	SELECT d.id, d.Department, d.parent, dCTE.lvl + 1 AS lvl,
		   cast(dCTE.TreePath 
                + ' -> ' 
                + d.Department AS varchar(8000)) as TreePath
	  FROM Departments AS d 
	 INNER JOIN DepartmentCTE AS dCTE ON d.parent = dCTE.id
	 WHERE d.parent IS NOT NULL
) 
SELECT TreePath
  FROM DepartmentCTE
 ORDER BY TreePath;

-- how about a path like a filesystem path
-- Replace this: + ' -> ' 
---With this: + '\' 




-- the TSQL REPLICATE function
SELECT REPLICATE('abc.', 3)


--only change is the final SELECT statement
;WITH DepartmentCTE AS
(
	SELECT id, Department, parent, 1 AS lvl,
           cast(Department as varchar(8000)) as TreePath 
  	  FROM Departments
	 WHERE parent IS NULL

	 UNION ALL 

	SELECT d.id, d.Department, d.parent, dCTE.lvl + 1 AS lvl,
		   cast(dCTE.TreePath 
                + ' -> ' 
                + d.Department AS varchar(8000)) as TreePath
	  FROM Departments AS d 
	 INNER JOIN DepartmentCTE AS dCTE ON d.parent = dCTE.id
	 WHERE d.parent IS NOT NULL
) 
SELECT REPLICATE('    ', lvl) 
     + Department as Department
  FROM DepartmentCTE
 ORDER BY TreePath;



--In the previous example it is necessary to include the ORDER BY 
--   TreePath to get all the departments to show up in the right 
--   order with sub-departments being shown under their parents. 



-- try this one
SELECT REPLICATE('.  ', lvl) 
     + Department as Department

-- or this
SELECT REPLICATE('   ', lvl) + '+'
     + Department as Department




-------------------------------------------
-- now for a royal family tree with CTEs
-------------------------------------------




use [cte_demo];  -- download the cte_demo database from http://stevestedman.com/?p=2839
--For more info on recursive CTE formatting see http://stevestedman.com/?p=1130
GO

-- take a look at the Royalty table
SELECT *
  FROM Royalty;







--If we wanted to create a query to trace Prince Williams maternal ancestors it would be 
-- straightforward with a CTE. We can see that Prince William has an id of 17 and a full 
-- name of William Arthur Phillip Windsor.
;WITH RoyalTreeCTE AS
(
   SELECT * 
     FROM Royalty 
	WHERE id = 17
    
	UNION ALL

   SELECT r.* 
     FROM Royalty AS r
    INNER JOIN RoyalTreeCTE AS rt ON rt.mother = r.id 
)
SELECT * 
 FROM RoyalTreeCTE;




 
--Finding the paternal ancestors would be very similar. Instead of using the mother 
--   column, use the father column.

;WITH RoyalTreeCTE AS
(
  SELECT * 
    FROM Royalty 
   WHERE id = 17

   UNION ALL

  SELECT r.*
    FROM Royalty AS r
   INNER JOIN RoyalTreeCTE AS rt ON rt.father = r.id 
)
SELECT * FROM RoyalTreeCTE;
  






--How can we write the query to find both the maternal and paternal sides of a family tree? 
-- Keep in mind that doing this will produce both the mother and father of each person in the tree.
;WITH RoyalTreeCTE AS
(
   SELECT * 
     FROM Royalty 
	WHERE id = 17

    UNION ALL

	--paternal recursive query
   SELECT r.* 
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.father = r.id

    UNION ALL

	-- maternal recursive query
   SELECT r.*
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.mother = r.id
)
SELECT * 
  FROM RoyalTreeCTE;


-- doesn't look much better than SELECT * FROM Royalty;




-- how to make it look better?
-- First we could apply the tree path method of formatting with the following query, 
--   also limits the results to only 3 levels of the hierarchy to fit on my screen:
;WITH RoyalTreeCTE AS
(
   SELECT *, 0 AS lvl,
          CAST(name AS VARCHAR(4096)) AS TreePath
     FROM Royalty WHERE id = 17  -- 17 is Prince William

    UNION ALL

   SELECT r.*, lvl + 1 AS lvl,
          CAST(rt.TreePath + '-->father:' +
               r.name AS VARCHAR(4096)) AS TreePath
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.father = r.id

    UNION ALL

   SELECT r.*, lvl + 1 AS lvl,
          CAST(rt.TreePath + '-->mother:' +
               r.name AS VARCHAR(4096)) AS TreePath
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.mother = r.id
)
SELECT TreePath
  FROM RoyalTreeCTE
 WHERE lvl < 3
 ORDER BY TreePath;








-- try this
SELECT REPLICATE('.      ', Level) + 
       Relationship + ': ' + name
-- and add a relationship column





;WITH RoyalTreeCTE AS
(
   SELECT *, 0 AS lvl,
          CAST(name AS VARCHAR(4096)) AS TreePath,
		  CAST('Top' AS VARCHAR(10)) AS Relationship
     FROM Royalty WHERE id = 17  -- 17 is Prince William
UNION ALL
   SELECT r.*, lvl + 1 AS lvl,
          CAST(rt.TreePath + '-->father:' +
               r.name AS VARCHAR(4096)) AS TreePath,
		  CAST('Father' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.father = r.id
UNION ALL
   SELECT r.*, lvl + 1 AS lvl,
          CAST(rt.TreePath + '-->mother:' +
               r.name AS VARCHAR(4096)) AS TreePath,
		  CAST('Mother' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.mother = r.id
)
SELECT REPLICATE('.      ', lvl) + 
       Relationship + ': ' + name
  FROM RoyalTreeCTE
 ORDER BY TreePath;






-- In order to improve the format we simply need to improve the formatting. We 
--  have all the right people, just not in the right order. 
-- step 1: add the Sorter column
;WITH RoyalTreeCTE AS
(
   SELECT *, 0 AS lvl,
          CAST('b' AS VARCHAR(4096)) AS Sorter,
          CAST('Top' AS VARCHAR(10)) AS Relationship
     FROM Royalty WHERE id = 17
UNION ALL
   SELECT r.*, lvl + 1 AS lvl,
          CAST(Sorter + 'a' AS VARCHAR(4096)) AS Sorter,
          CAST('Father' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.father = r.id
UNION ALL
   SELECT r.*, lvl + 1 as lvl,
          cast(Sorter + 'c' AS VARCHAR(4096)) AS Sorter,
          cast('Mother' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.mother = r.id
)
SELECT sorter, name
  FROM RoyalTreeCTE
 ORDER BY Sorter + 'bbbbbbbbbbbbb';







--next add the indentation
--  SELECT REPLICATE('.           ', Level) + 
--     Relationship + ': ' + name

;WITH RoyalTreeCTE AS
(
   SELECT *, 0 AS lvl,
          CAST('b' AS VARCHAR(4096)) AS Sorter,
          CAST('Top' AS VARCHAR(10)) AS Relationship
     FROM Royalty WHERE id = 17
UNION ALL
   SELECT r.*, lvl + 1 AS lvl,
          CAST(Sorter + 'a' AS VARCHAR(4096)) AS Sorter,
          CAST('Father' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.father = r.id
UNION ALL
   SELECT r.*, lvl + 1 as lvl,
          cast(Sorter + 'c' AS VARCHAR(4096)) AS Sorter,
          cast('Mother' AS VARCHAR(10)) AS Relationship
     FROM Royalty r
    INNER JOIN RoyalTreeCTE rt ON rt.mother = r.id
)
SELECT REPLICATE('.           ', lvl) + 
       Relationship + ': ' + name
  FROM RoyalTreeCTE
 ORDER BY Sorter + 'bbbbbbbbbbbbb';

